package com.engibank.middleware;

import static com.engibank.util.Soap.extractWithKey;
import static com.engibank.util.Soap.genErrorReply;

import com.engibank.model.User;
import com.engibank.util.Database;
import com.warrenstrange.googleauth.GoogleAuthenticator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.EntityManager;
import javax.persistence.TypedQuery;
import javax.xml.namespace.QName;
import javax.xml.soap.Node;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;


public class AuthToken implements SOAPHandler<SOAPMessageContext> {

  private static final String AUTHKEY = "authToken";
  private static final Logger LOGGER = Logger.getLogger(AuthToken.class.getName());

  @Override
  public Set<QName> getHeaders() {
    Set<QName> header = new HashSet<>();
    header.add(new QName(AUTHKEY));
    return header;
  }

  @Override
  public boolean handleMessage(SOAPMessageContext soapMessageContext) {
    boolean outRequest = (boolean) soapMessageContext.get(MessageContext.MESSAGE_OUTBOUND_PROPERTY);

    if (!outRequest) {
      LOGGER.info("Entering token auth handler");

      SOAPMessage message = soapMessageContext.getMessage();

      SOAPHeader header = null;

      try {
        header = message.getSOAPHeader();
      } catch (SOAPException e) {
        LOGGER.log(Level.SEVERE, "Invalid header", e);
        genErrorReply(message, "Invalid header");
      }

      Iterator<Node> headerIterator = header.getChildElements();

      Node authNode = extractWithKey(headerIterator, AUTHKEY);

      if (authNode == null) {
        genErrorReply(message, "No auth");
      }

      SOAPElement authElement = (SOAPElement) authNode;

      Iterator<Node> authChildIterator = authElement.getChildElements();
      Node accountNode = extractWithKey(authChildIterator, "account");

      authChildIterator = authElement.getChildElements();
      Node tokenNode = extractWithKey(authChildIterator, "token");

      if (accountNode == null || tokenNode == null) {
        genErrorReply(message, "No account and/or token provided");
      }

      assert accountNode != null;
      assert tokenNode != null;

      String accountNum = accountNode.getValue();
      String token = tokenNode.getValue();

      LOGGER.info(String.format("Detected username and token: %s %s", accountNum, token));

      Database database = new Database();
      EntityManager em = database.getEntityManager();

      TypedQuery<User> query = em.createQuery(
          "SELECT u FROM User u WHERE u.accountNumber = :acc_num", User.class);
      query.setParameter("acc_num", accountNum);

      User authUser = null;

      try {
        authUser = query.getSingleResult();
      } catch (Exception e) {
        LOGGER.log(Level.SEVERE, "User query fail", e);
        genErrorReply(message, "Invalid account and/or token");
      } finally {
        database.close();
      }

      if (authUser == null) {
        genErrorReply(message, "Invalid account and/or token");
      }

      assert authUser != null;

      GoogleAuthenticator googleAuth = new GoogleAuthenticator();
      boolean isCodeValid = googleAuth.authorize(authUser.getSecret(), Integer.parseInt(token));

      if (!isCodeValid) {
        genErrorReply(message, "Invalid account and/or token");
      }

      LOGGER.info(String.format("Auth user as %s (%s)", accountNum, authUser.getName()));
      soapMessageContext.put("user", authUser);
      soapMessageContext.setScope("user", MessageContext.Scope.APPLICATION);
    }
    return true;
  }

  @Override
  public boolean handleFault(SOAPMessageContext soapMessageContext) {
    return false;
  }

  @Override
  public void close(MessageContext messageContext) {

  }
}
