package com.engibank.service;

import com.engibank.model.Account;
import com.engibank.model.User;
import com.engibank.util.Database;
import com.engibank.util.Jwt;
import com.warrenstrange.googleauth.GoogleAuthenticator;
import java.util.logging.Logger;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.persistence.EntityManager;
import javax.persistence.TypedQuery;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.ws.Holder;

@WebService()
public class AuthService {
  private static final Logger LOGGER = Logger.getLogger(AuthService.class.getName());

  /**
   * Login endpoint.
   *
   * @return jwt key
   */
  @WebMethod
  public String login(
      @XmlElement(required = true)
      @WebParam(name = "accountNumber", mode = WebParam.Mode.IN) String accountNumber,
      @XmlElement(required = true)
      @WebParam(name = "token", mode = WebParam.Mode.IN) String token) throws Exception {
    Database database = new Database();
    EntityManager em = database.getEntityManager();

    TypedQuery<User> query = em.createQuery(
        "SELECT o FROM Account a INNER JOIN a.owner o WHERE a.id = :acc AND not a.isVirtual",
        User.class);
    query.setParameter("acc", Integer.parseInt(accountNumber));

    User authUser = null;

    try {
      authUser = query.getSingleResult();
    } catch (Exception e) {
      LOGGER.severe(e.toString());
      throw new Exception("Invalid account and/or token");
    } finally {
      database.close();
    }

    if (authUser == null) {
      throw new Exception("Invalid account and/or token");
    }

    GoogleAuthenticator googleAuth = new GoogleAuthenticator();
    boolean isCodeValid = googleAuth.authorize(authUser.getSecret(), Integer.parseInt(token));

    if (!isCodeValid) {
      throw new Exception("Invalid account and/or token");
    }

    LOGGER.info(String.format("Auth user as %s (%s)", accountNumber, authUser.getName()));
    return Jwt.generateJwt(authUser.getId(), authUser.isApi());
  }

  /**
   * Register endpoint.
   *
   * @param name user name
   * @param bank bank name
   */
  @WebMethod
  public void register(
      @WebParam(name = "user", mode = WebParam.Mode.OUT) Holder<User> user,
      @WebParam(name = "account", mode = WebParam.Mode.OUT) Holder<Account> account,
      @XmlElement(required = true)
      @WebParam(name = "name", mode = WebParam.Mode.IN) String name,
      @XmlElement(required = true)
      @WebParam(name = "bank", mode = WebParam.Mode.IN) String bank) {

    Database database = new Database();
    EntityManager em = database.getEntityManager();

    em.getTransaction().begin();

    User tempUser = new User(name, bank, 0);
    Account tempAcc = new Account(tempUser, false);

    em.persist(tempUser);
    em.persist(tempAcc);

    em.getTransaction().commit();

    user.value = tempUser;
    account.value = tempAcc;

    database.close();
  }
}
