package com.engibank.service;

import com.engibank.model.Account;
import com.engibank.model.Transaction;
import com.engibank.model.User;
import com.engibank.util.Database;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.annotation.Resource;
import javax.jws.HandlerChain;
import javax.jws.WebMethod;
import javax.jws.WebParam;
import javax.jws.WebService;
import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.TypedQuery;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.MessageContext;

@HandlerChain(file = "../authKey.xml")
@WebService()
public class TransferService {
  private static final Logger LOGGER = Logger.getLogger(TransferService.class.getName());

  @Resource
  WebServiceContext context;

  /**
   * Generate new transaction.
   * by validate if account number exists
   * and if account balance is enough for transfer
   *
   * @return new is transaction created
   */

  @WebMethod
  public boolean transfer(
      @XmlElement(required = true)
      @WebParam(name = "accountTujuan", mode = WebParam.Mode.IN) String accTujuan,
      @XmlElement(required = true)
      @WebParam(name = "jumlahTransfer", mode = WebParam.Mode.IN) int jlhTransfer) 
      throws Exception {

    LOGGER.log(Level.FINE, "New Transfer request");
    MessageContext ctx = context.getMessageContext();
    User loggedinUser = (User) ctx.get("user");

    Database database = new Database();
    EntityManager em = database.getEntityManager();

    em.getTransaction().begin();

    Account accPengirim = loggedinUser.getPrimaryAccount();
    boolean saldoCukup = false;
    if (accPengirim.getBalance() > jlhTransfer) {
      saldoCukup = true;
    } else {
      throw new Exception("Saldo tidak cukup");
    }

    Account accPenerima = null;
    int accNumTujuan = 0;
    if (accTujuan.length() == 12) {
      accNumTujuan = Integer.parseInt(accTujuan);
    } else if (accTujuan.length() == 24) {
      accNumTujuan = Integer.parseInt(accTujuan.substring(12, 24));
    }

    TypedQuery<Account> query = em.createQuery(
        "SELECT a FROM Account a JOIN User u WHERE a.id = :acc_num and u.id = :user_id",
        Account.class);
    query.setParameter("acc_num", accNumTujuan);
    query.setParameter("user_id", loggedinUser.getId());

    try {
      accPenerima = query.getSingleResult();
    } catch (NoResultException e) {
      LOGGER.log(Level.INFO, "No account found");
      return false;
    }

    //proses tf
    if (saldoCukup && accPenerima != null) {
      //pengirim
      int saldoFinalPengirim = accPengirim.getBalance() - jlhTransfer;
      accPengirim.setBalance(saldoFinalPengirim);
      em.merge(accPengirim);

      //penerima
      int saldoFinalPenerima = accPenerima.getBalance() + jlhTransfer;
      accPenerima.setBalance(saldoFinalPenerima);


      //create new transaction
      Transaction t = new Transaction(accPengirim, accPenerima, jlhTransfer);
      em.persist(t);

      User u = null;
      Account accPrimaryPenerima = null;
      //check if virtual or not the tujuan acc
      if (accPenerima.isVirtual()) {
        u = accPenerima.getOwner();
        accPrimaryPenerima = u.getPrimaryAccount();
        accPrimaryPenerima.setBalance(accPrimaryPenerima.getBalance() + jlhTransfer);
        em.merge(accPrimaryPenerima);
      }

      em.getTransaction().commit();
      database.close();

      LOGGER.log(Level.FINE, String.format("New transfer from %s to %s with amount of %d",
          accPengirim.getAccountNumber(),
          accPenerima.getAccountNumber(),
          jlhTransfer));
    }
    return true;
  }

  /**
   * Get transaction from current user.
   *
   * @return List of transaction
   */

  @WebMethod
  public List<Transaction> getTransactionHistory() {
    LOGGER.log(Level.FINE, "Transaction request");

    MessageContext ctx = context.getMessageContext();
    User loggedinUser = (User) ctx.get("user");
    List<Account> listOfAccount = loggedinUser.getAccount();

    Database database = new Database();
    EntityManager em = database.getEntityManager();

    TypedQuery<Transaction> query = em.createQuery(
        "SELECT t FROM Transaction t WHERE t.from in :accountNumber OR t.to in :accountNumber",
        Transaction.class);
    query.setParameter("accountNumber", listOfAccount);
    List<Transaction> temp = query.getResultList();
    if (temp.size() != 0) {
      return temp;
    } else {
      return null;
    }
  }

  /**
   * Check if transaction exists.
   *
   * @return boolean
   */

  @WebMethod
  public boolean checkTransaction(
      @XmlElement(required = true)
      @WebParam(name = "accountTujuan", mode = WebParam.Mode.IN) String accTujuan,
      @XmlElement(required = true)
      @WebParam(name = "jumlahTransfer", mode = WebParam.Mode.IN) int jlhTransfer,
      @XmlElement(required = true)
      @WebParam(name = "startTime", mode = WebParam.Mode.IN) String start,
      @XmlElement(required = true)
      @WebParam(name = "endTime", mode = WebParam.Mode.IN) String end)
      throws Exception {

    LOGGER.log(Level.FINE, "Check transaction request");

    Database database = new Database();
    EntityManager em = database.getEntityManager();

    int accNumTujuan = 0;
    try {
      if (accTujuan.length() == 12) {
        accNumTujuan = Integer.parseInt(accTujuan);
      } else if (accTujuan.length() == 24) {
        accNumTujuan = Integer.parseInt(accTujuan.substring(12, 24));
      }
    } catch (Exception e) {
      throw new Exception("Invalid account number");
    }


    Date startTime = null;
    Date endTime = null;
    SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    try {
      startTime = dateFormat.parse(start);
      endTime = dateFormat.parse(end);
    } catch (ParseException e) {
      LOGGER.log(Level.INFO, "False string format");
      throw new Exception("Invalid datetime format, expected yyyy-MM-dd HH:mm:ss");
    }

    MessageContext ctx = context.getMessageContext();
    Account accPenerima = null;
    TypedQuery<Account> query = em.createQuery(
        "SELECT a FROM Account a JOIN User u WHERE a.id = :acc_num",
        Account.class);
    query.setParameter("acc_num", accNumTujuan);

    try {
      accPenerima = query.getSingleResult();
    } catch (NoResultException e) {
      LOGGER.log(Level.INFO, "No account found");
      return false;
    }

    User loggedinUser = (User) ctx.get("user");

    if (accPenerima.getOwner().getId() != loggedinUser.getId()) {
      throw new Exception("Account is not owned by you");
    }
    TypedQuery<Transaction> queryTrans = em.createQuery(
        "SELECT t FROM Transaction t WHERE t.to = :to "
            + "and t.amount = :amount and t.timestamp BETWEEN :start and :endtime",
        Transaction.class);

    queryTrans.setParameter("to", accPenerima);
    queryTrans.setParameter("amount", jlhTransfer);
    queryTrans.setParameter("start", startTime);
    queryTrans.setParameter("endtime", endTime);

    List<Transaction> t = null;
    try {
      t = queryTrans.getResultList();
    } catch (NoResultException e) {
      LOGGER.log(Level.INFO, "No transaction found");
    } finally {
      database.close();
    }

    return t != null && t.size() != 0;
  }
}
